import pickle
from pathlib import Path
from typing import Any, ClassVar

from .base import FileType


class PickleFile(FileType):
    """A class for validating and writing JSON files.

    Examples:
        >>> with tempfile.NamedTemporaryFile() as tmp_file:
        ...     fp = Path(tmp_file.name)
        ...     PickleFile.write(fp, {"key": "value"})
        ...     with open(fp, 'rb') as f:
        ...         print(pickle.load(f))
        {'key': 'value'}

    `PickleFile` can support any type that can be pickled:

        >>> PickleFile.validate("Hello!")
        >>> PickleFile.validate(123)
        >>> PickleFile.validate({"key": "value"})
        >>> PickleFile.validate([1, 2, 3])
        >>> PickleFile.validate({1, 2, 3})
        >>> PickleFile.validate(lambda x: x)
        Traceback (most recent call last):
            ...
        _pickle.PicklingError: Can't pickle <function <lambda> at ...>: ...
    """

    extension: ClassVar[str] = ".pkl"

    @classmethod
    def validate(cls, contents: Any):
        pickle.dumps(contents)

    @classmethod
    def write(cls, file_path: Path, contents: Any) -> None:
        with open(file_path, "wb") as f:
            pickle.dump(contents, f)



import csv
import io
from pathlib import Path
from typing import Any, ClassVar

from .base import FileType


class CSVFile(FileType):
    """A class for validating and writing comma-separated-value (CSV) files in one of several representations.

    Examples:
        >>> str_data = '''
        ... Name, Age, City
        ... "Alice Smith", 30, New York
        ... "Bob Jones, Jr.", 25, Los Angeles
        ... '''
        >>> with tempfile.NamedTemporaryFile() as tmp_file:
        ...     CSVFile.write(Path(tmp_file.name), str_data)
        ...     print(Path(tmp_file.name).read_text())
        Name, Age, City
        "Alice Smith", 30, New York
        "Bob Jones, Jr.", 25, Los Angeles

    For string formatted data, errors are raised if any of the CSV rows have inconsistent fields:

        >>> str_data_invalid = '''
        ... Name, Age, City
        ... Alice, 30, New York, 43
        ... Bob, 25, Los Angeles
        ... '''
        >>> CSVFile.validate(str_data_invalid)
        Traceback (most recent call last):
            ...
        ValueError: A string CSV can't have inconsistent fields in rows; got
            Row 0 has extra columns; got ' 43'

    CSV files can also accommodate a dictionary of column names to column values. Note that as raw values are
    used here, the CSV file will not have the spaces in it the prior stringified version did, for example.

        >>> column_map_data = {
        ...     "Name": ["Alice Smith", "Bob Jones, Jr."],
        ...     "Age": [30, 25],
        ...     "City": ["New York", "Los Angeles"]
        ... }
        >>> with tempfile.NamedTemporaryFile() as tmp_file:
        ...     CSVFile.write(Path(tmp_file.name), column_map_data)
        ...     print(Path(tmp_file.name).read_text().strip())
        Name,Age,City
        Alice Smith,30,New York
        "Bob Jones, Jr.",25,Los Angeles

    In this format, we must (1) have all column names be strings:

        >>> column_map_data_invalid = {
        ...     "Name": ["Alice Smith", "Bob Jones, Jr."],
        ...     1: [30, 25],
        ...     "City": ["New York", "Los Angeles"]
        ... }
        >>> CSVFile.validate(column_map_data_invalid)
        Traceback (most recent call last):
            ...
        ValueError: Column-maps must have all string keys; got 1 (int)

    (2) Have all column values be lists of entries

        >>> column_map_data_invalid = {
        ...     "Name": ["Alice Smith", "Bob Jones, Jr."],
        ...     "Age": (30, 25),
        ...     "City": ["New York", "Los Angeles"]
        ... }
        >>> CSVFile.validate(column_map_data_invalid)
        Traceback (most recent call last):
            ...
        ValueError: Column-maps must have all list values; got cols Age (tuple)

    and (3) have all column values be lists of the same length:

        >>> column_map_data_invalid = {
        ...     "Name": ["Alice Smith", "Bob Jones, Jr."],
        ...     "Age": [30, 25, 40],
        ...     "City": ["New York", "Los Angeles"]
        ... }
        >>> CSVFile.validate(column_map_data_invalid)
        Traceback (most recent call last):
            ...
        ValueError: Column-maps must have all lists of the same length 2; got Age (3)

    Another accepted format is a list of rows mapping column name to row value:

        >>> list_rowmap_data = [
        ...     {"Name": "Alice", "Age": 30, "City": "New York"},
        ...     {"Name": "Bob", "Age": 25, "City": "Los Angeles"}
        ... ]
        >>> with tempfile.NamedTemporaryFile() as tmp_file:
        ...     CSVFile.write(Path(tmp_file.name), list_rowmap_data)
        ...     print(Path(tmp_file.name).read_text().strip())
        Name,Age,City
        Alice,30,New York
        Bob,25,Los Angeles

    In this format, we must have all rows have the same set of keys:

        >>> list_rowmap_data_invalid = [
        ...     {"Name": "Alice", "Age": 30, "City": "New York"},
        ...     {"Name": "Bob", "City": "Los Angeles", "Name2": "Bob2"},
        ... ]
        >>> CSVFile.validate(list_rowmap_data_invalid)
        Traceback (most recent call last):
            ...
        ValueError: Row-maps must be consistent; got
            Row 1 has extra keys ['Name2'], missing keys ['Age']

    Finally, we can also pass a list of rows with the first row being the header:

        >>> list_rows_data = [
        ...     ["Name", "Age", "City"],
        ...     ["Alice", 30, "New York"],
        ...     ["Bob", 25, "Los Angeles"]
        ... ]
        >>> with tempfile.NamedTemporaryFile() as tmp_file:
        ...     CSVFile.write(Path(tmp_file.name), list_rows_data)
        ...     print(Path(tmp_file.name).read_text().strip())
        Name,Age,City
        Alice,30,New York
        Bob,25,Los Angeles

    In this format, we must have the first row be a list of strings and all subsequent rows be lists of the
    same length:

        >>> list_rows_data_invalid = [
        ...     ["Name", "Age", 3],
        ...     ["Alice", 30, "New York"],
        ...     ["Bob", 25, "Los Angeles"]
        ... ]
        >>> CSVFile.validate(list_rows_data_invalid)
        Traceback (most recent call last):
            ...
        ValueError: A list of lines must have the first row as str headers; got ['Name', 'Age', 3]
        >>> list_rows_data_invalid = [
        ...     ["Name", "Age", "City"],
        ...     ["Alice", 30, "New York", 32],
        ...     ["Bob", 25, "Los Angeles"]
        ... ]
        >>> CSVFile.validate(list_rows_data_invalid)
        Traceback (most recent call last):
            ...
        ValueError: Rows must be consistent; got
            Row 0 has 4 columns; expected 3

    In any format, if the data is empty, it will yield an empty CSV file:
        >>> CSVFile._parse("")
        ''
        >>> CSVFile._parse({})
        ''
        >>> CSVFile._parse([])
        ''

    If the data is in none of these formats, it will not be accepted:

        >>> CSVFile.validate(123)
        Traceback (most recent call last):
            ...
        TypeError: Contents must be a string, dict, or list; got <class 'int'>
    """

    extension: ClassVar[str] = ".csv"
    separator: ClassVar[str] = ","

    @classmethod
    def validate(cls, contents: Any):
        cls._parse(contents)

    @classmethod
    def _parse_str(cls, contents: str) -> str:
        """Validates a string is a valid CSV file contents and returns it.

        Also strips leading and trailing whitespace from each line.
        """

        contents = "\n".join(line.strip() for line in contents.splitlines())

        data = csv.DictReader(contents.strip().splitlines(), delimiter=cls.separator, restkey="err")

        bad_rows = []
        for i, row in enumerate(data):
            if row.get("err", None) is not None:
                bad_rows.append(f"Row {i} has extra columns; got '{cls.separator.join(row['err'])}'")

        if bad_rows:
            bad_rows_str = "\n".join(bad_rows)
            raise ValueError(f"A string CSV can't have inconsistent fields in rows; got\n{bad_rows_str}")
        return contents.strip()

    @classmethod
    def __dict_rows_to_str(cls, fieldnames: list[str], rows: list[dict]) -> str:
        str_io = io.StringIO()
        writer = csv.DictWriter(str_io, fieldnames=fieldnames, delimiter=cls.separator)

        writer.writeheader()
        for row in rows:
            writer.writerow(row)

        return str_io.getvalue()

    @classmethod
    def _parse_dict(cls, contents: dict[str, list[Any]]) -> str:
        """Parses a dictionary of column names to column values into a CSV string."""

        bad_keys = [f"{k} ({type(k).__name__})" for k in contents if not isinstance(k, str)]
        if bad_keys:
            raise ValueError(f"Column-maps must have all string keys; got {', '.join(bad_keys)}")

        bad_keys = [f"{k} ({type(v).__name__})" for k, v in contents.items() if not isinstance(v, list)]
        if bad_keys:
            raise ValueError(f"Column-maps must have all list values; got cols {', '.join(bad_keys)}")

        N = len(next(iter(contents.values())))
        bad_keys = [f"{k} ({len(v)})" for k, v in contents.items() if len(v) != N]
        if bad_keys:
            bad_keys_str = ", ".join(bad_keys)
            raise ValueError(f"Column-maps must have all lists of the same length {N}; got {bad_keys_str}")

        fieldnames = list(contents.keys())
        rows = [dict(zip(fieldnames, row, strict=False)) for row in zip(*contents.values(), strict=False)]
        return cls.__dict_rows_to_str(fieldnames, rows)

    @classmethod
    def _parse_list_rows(cls, contents: list[dict[str, Any]]) -> str:
        """Parses a list of rows (as dictionaries) into a CSV string."""

        bad_keys = [k for k in contents[0] if not isinstance(k, str)]
        if bad_keys:
            raise ValueError(f"Column-maps must have all string keys; got {bad_keys}")

        fieldnames = list(contents[0].keys())

        bad_rows = []
        for i, row in enumerate(contents):
            missing_keys = set(fieldnames) - set(row.keys())
            extra_keys = set(row.keys()) - set(fieldnames)

            err_parts = []
            if extra_keys:
                err_parts.append(f"extra keys {sorted(extra_keys)}")
            if missing_keys:
                err_parts.append(f"missing keys {sorted(missing_keys)}")

            if err_parts:
                bad_rows.append(f"Row {i} has {', '.join(err_parts)}")

        if bad_rows:
            bad_rows_str = "\n".join(bad_rows)
            raise ValueError(f"Row-maps must be consistent; got\n{bad_rows_str}")

        return cls.__dict_rows_to_str(fieldnames, contents)

    @classmethod
    def _parse_list_lines(cls, contents: list[list[Any]]) -> str:
        """Parses a list of rows (as lists) into a CSV string, with the first row being the header."""

        fieldnames = contents[0]
        if not all(isinstance(item, str) for item in fieldnames):
            raise ValueError(f"A list of lines must have the first row as str headers; got {fieldnames}")

        N_cols = len(fieldnames)

        bad_rows = []
        rows_as_dicts = []
        for i, row in enumerate(contents[1:]):
            if not isinstance(row, list):
                bad_rows.append(f"Row {i} is not a list; got {type(row)}")
            if len(row) != N_cols:
                bad_rows.append(f"Row {i} has {len(row)} columns; expected {N_cols}")
            rows_as_dicts.append(dict(zip(fieldnames, row, strict=False)))

        if bad_rows:
            bad_rows_str = "\n".join(bad_rows)
            raise ValueError(f"Rows must be consistent; got\n{bad_rows_str}")

        return cls.__dict_rows_to_str(fieldnames, rows_as_dicts)

    @classmethod
    def _parse(cls, contents: Any) -> str:
        match contents:
            case str():
                return cls._parse_str(contents)
            case dict():
                return cls._parse_dict(contents) if contents else ""
            case list() if all(isinstance(item, dict) for item in contents):
                return cls._parse_list_rows(contents) if contents else ""
            case list() if all(isinstance(item, list) for item in contents):
                return cls._parse_list_lines(contents) if contents else ""
            case _:
                raise TypeError(f"Contents must be a string, dict, or list; got {type(contents)}")

    @classmethod
    def write(cls, file_path: Path, contents: Any) -> None:
        file_path.write_text(cls._parse(contents))



from typing import Any

import pytest

from .yaml_to_disk import yaml_disk


@pytest.fixture(autouse=True)
def ___yaml_to_disk_add_doctest(doctest_namespace: dict[str, Any]) -> None:
    doctest_namespace.update({"yaml_disk": yaml_disk})



from __future__ import annotations

import logging
import re
import tempfile
from dataclasses import dataclass
from pathlib import Path
from typing import Any, ClassVar

from yaml import load

from .file_matchers import FILE_TYPE_MATCHER

try:
    from yaml import CLoader as Loader
except ImportError:
    from yaml import Loader

logger = logging.getLogger(__name__)

DISK_CONTENTS_T = str | Path | dict | list


@dataclass
class File:
    rel_path: Path
    contents: Any

    def write(
        self,
        root_dir: Path,
        do_overwrite: bool = False,
        use_txt_on_unk_str_files: bool = True,
    ):
        """Write the file to the specified directory."""

        file_path = root_dir / self.rel_path
        if file_path.exists():
            if not do_overwrite:
                raise FileExistsError(f"File already exists: {file_path}")
            elif file_path.is_dir():
                raise IsADirectoryError(f"Path exists and is a directory: {file_path}")
            else:
                file_path.unlink()

        file_path.parent.mkdir(parents=True, exist_ok=True)

        if self.contents is None:
            file_path.touch()
            return

        try:
            file_type = FILE_TYPE_MATCHER(self.rel_path)
        except ValueError:
            if isinstance(self.contents, str) and use_txt_on_unk_str_files:
                file_type = FILE_TYPE_MATCHER(".txt")
            else:
                raise

        try:
            file_type.validate(self.contents)
        except Exception as e:
            raise ValueError(f"Contents for {file_path} fail validation under {file_type.__name__}") from e

        file_type.write(file_path, self.contents)


@dataclass
class Directory:
    rel_path: Path
    contents: list[File | Directory]

    def write(self, root_dir: Path, **kwargs):
        """Write the directory and its contents to the specified directory."""

        new_root = root_dir / self.rel_path
        if new_root.exists() and new_root.is_file():
            raise NotADirectoryError(f"Path exists and is a file: {new_root}")

        for item in self.contents:
            item.write(new_root, **kwargs)


class YamlDisk:
    """A class to represent a collection of files in a root directory in a YAML or YAML-derived format.

    This class is a context manager and a functor, and can be used in either mode. It preserves no state
    beyond that required for context management. The singleton object released below should generally be used,
    rather than creating new instances.

    If this package is installed as a plugin, it will automatically be available as a doctest fixture.
    """

    _temp_dir: tempfile.TemporaryDirectory | None
    _FILENAME_REGEX: ClassVar[re.Pattern] = re.compile(r".*\.(yaml|yml)$")
    _USE_TXT_ON_UNK_STR_FILES: ClassVar[bool] = True

    def __init__(self):
        self._temp_dir = None

    def __enter__(self):
        # Create temporary directory
        self._temp_dir = tempfile.TemporaryDirectory()
        return self(self.disk_contents, root_dir=Path(self._temp_dir.name), **self._write_kwargs)

    def __exit__(self, exc_type, exc_val, exc_tb):
        # Clean up temporary directory
        self._temp_dir.cleanup()

    def __call__(
        self,
        disk_contents: DISK_CONTENTS_T,
        root_dir: str | Path | None = None,
        do_overwrite: bool = False,
        **kwargs,
    ) -> Path | YamlDisk:
        if root_dir is None:
            # Here, we're just preparing for context manager mode -- not actually writing anything.
            self.disk_contents = disk_contents
            self._write_kwargs = kwargs
            return self

        match root_dir:
            case str():
                root_dir = Path(root_dir)
            case Path():
                pass
            case _:
                raise TypeError(f"root_dir must be a string or Path, got {type(root_dir)}")

        if not root_dir.exists():
            raise FileNotFoundError(f"root_dir does not exist: {root_dir}")
        elif not root_dir.is_dir():
            raise NotADirectoryError(f"root_dir is not a directory: {root_dir}")

        parsed_contents = self._parse(disk_contents)
        self._write(parsed_contents, root_dir, do_overwrite=do_overwrite, **kwargs)
        return root_dir

    @classmethod
    def _is_yaml_file(cls, path: str) -> bool:
        """Check if the given string represents a YAML file."""

        if not isinstance(path, str):
            raise TypeError(f"Expected a string, got {type(path)}")

        return cls._FILENAME_REGEX.match(path) is not None

    @classmethod
    def _read_yaml_file(cls, yaml_fp: Path) -> dict | list:
        """Read a YAML file and return its contents."""

        if not yaml_fp.is_file():
            raise FileNotFoundError(f"File not found: {yaml_fp}")
        return cls._read_yaml_str(yaml_fp.read_text())

    @classmethod
    def _read_yaml_str(cls, yaml_str: str) -> dict | list:
        """Load a YAML string and return its contents."""

        return load(yaml_str, Loader=Loader)

    @classmethod
    def _parse_yaml_contents(cls, yaml_contents: dict | list) -> list[File | Directory]:
        """Parse the YAML contents and return a list of File or Directory objects."""

        match yaml_contents:
            case dict() as yaml_dict:
                yaml_list = [({k: v} if v else k) for k, v in yaml_dict.items()]
            case list():
                yaml_list = yaml_contents
            case _:
                raise TypeError(f"YAML contents must be a dictionary or list; got {type(yaml_contents)}")

        parsed_contents = []
        for item in yaml_list:
            match item:
                case str() as dir_name if item.endswith("/"):
                    parsed_contents.append(Directory(rel_path=Path(dir_name), contents=[]))
                case str() as file_name:
                    parsed_contents.append(File(rel_path=Path(file_name), contents=None))
                case dict() if len(item) == 1 and isinstance(next(iter(item.keys())), str):
                    name = next(iter(item.keys()))
                    contents = item[name]

                    name = Path(name)

                    if name.suffix != "/" and name.suffix:
                        parsed_contents.append(File(rel_path=name, contents=contents))
                    else:
                        nested_contents = cls._parse_yaml_contents(contents)
                        parsed_contents.append(Directory(rel_path=name, contents=nested_contents))

                case dict():
                    raise ValueError(f"Invalid format: {item}. Expected a single key (str) - value pair.")
                case _:
                    raise TypeError(f"Invalid item type: {type(item)}. Expected str or dict.")

        return parsed_contents

    @classmethod
    def _parse(cls, disk_contents: DISK_CONTENTS_T) -> list[File | Directory]:
        match disk_contents:
            case str() as yaml_fp_str if cls._is_yaml_file(yaml_fp_str):
                logging.debug(f"Inferring passed string {yaml_fp_str} is a file path")
                yaml_contents = cls._read_yaml_file(Path(yaml_fp_str))
            case Path() as yaml_fp:
                yaml_contents = cls._read_yaml_file(yaml_fp)
            case str() as yaml_str:
                yaml_contents = cls._read_yaml_str(yaml_str)
            case dict() | list():
                yaml_contents = disk_contents
            case _:
                raise TypeError(
                    f"disk_contents must be a string, Path, dict, or list; got {type(disk_contents)}"
                )

        return cls._parse_yaml_contents(yaml_contents)

    @classmethod
    def _write(
        cls,
        parsed_contents: list[File | Directory],
        root_dir: Path,
        do_overwrite: bool = False,
        **kwargs,
    ):
        base = Directory(rel_path=Path(), contents=parsed_contents)

        if "use_txt_on_unk_str_files" not in kwargs:
            kwargs["use_txt_on_unk_str_files"] = cls._USE_TXT_ON_UNK_STR_FILES

        base.write(root_dir, do_overwrite=do_overwrite, **kwargs)


yaml_disk = YamlDisk()



from pathlib import Path
from typing import Any, ClassVar

from yaml import dump

from .base import FileType

try:
    from yaml import CDumper as Dumper
except ImportError:
    from yaml import Dumper


class YAMLFile(FileType):
    """A class for validating and writing JSON files.

    Examples:
        >>> with tempfile.NamedTemporaryFile() as tmp_file:
        ...     fp = Path(tmp_file.name)
        ...     YAMLFile.write(fp, {"key": "value"})
        ...     print(fp.read_text().strip())
        key: value

    `YAMLFile` can accept a variety of data types, including lambda functions.

        >>> YAMLFile.validate("Hello!")
        >>> YAMLFile.validate(123)
        >>> YAMLFile.validate({"key": "value"})
        >>> YAMLFile.validate([1, 2, 3])
        >>> YAMLFile.validate({1, 2, 3})
        >>> YAMLFile.validate(lambda x: x)
    """

    extension: ClassVar[frozenset[str]] = frozenset({".yaml", ".yml"})

    @classmethod
    def validate(cls, contents: Any):
        dump(contents, Dumper=Dumper)

    @classmethod
    def write(cls, file_path: Path, contents: Any) -> None:
        file_path.write_text(dump(contents, Dumper=Dumper))



"""Test set-up and fixtures code."""

import importlib
import tempfile
from pathlib import Path
from typing import Any

import pytest

import yaml_to_disk
import yaml_to_disk.file_matchers
import yaml_to_disk.file_types
import yaml_to_disk.file_types.base
import yaml_to_disk.file_types.csv
import yaml_to_disk.file_types.json
import yaml_to_disk.file_types.pkl
import yaml_to_disk.file_types.txt
import yaml_to_disk.file_types.yaml
import yaml_to_disk.pytest_plugin
import yaml_to_disk.yaml_to_disk

importlib.reload(yaml_to_disk)
importlib.reload(yaml_to_disk.pytest_plugin)
importlib.reload(yaml_to_disk.yaml_to_disk)
importlib.reload(yaml_to_disk.file_matchers)
importlib.reload(yaml_to_disk.file_types)
importlib.reload(yaml_to_disk.file_types.base)
importlib.reload(yaml_to_disk.file_types.csv)
importlib.reload(yaml_to_disk.file_types.json)
importlib.reload(yaml_to_disk.file_types.yaml)
importlib.reload(yaml_to_disk.file_types.pkl)
importlib.reload(yaml_to_disk.file_types.txt)


@pytest.fixture(autouse=True)
def __setup_doctest_namespace(doctest_namespace: dict[str, Any]):
    doctest_namespace.update({"tempfile": tempfile, "Path": Path})



import json
from pathlib import Path
from typing import Any, ClassVar

from .base import FileType


class JSONFile(FileType):
    """A class for validating and writing JSON files.

    Examples:
        >>> with tempfile.NamedTemporaryFile() as tmp_file:
        ...     fp = Path(tmp_file.name)
        ...     JSONFile.write(fp, {"key": "value"})
        ...     fp.read_text()
        '{"key": "value"}'

    `JSONFile` can accept any type that is JSON serializable.

        >>> JSONFile.validate("Hello!")
        >>> JSONFile.validate(123)
        >>> JSONFile.validate({"key": "value"})
        >>> JSONFile.validate([1, 2, 3])
        >>> JSONFile.validate({1, 2, 3})
        Traceback (most recent call last):
            ...
        TypeError: Object of type set is not JSON serializable
    """

    extension: ClassVar[str] = ".json"

    @classmethod
    def validate(cls, contents: Any):
        json.dumps(contents)

    @classmethod
    def write(cls, file_path: Path, contents: Any) -> None:
        file_path.write_text(json.dumps(contents))



import abc
from pathlib import Path
from typing import Any, ClassVar


class FileType(abc.ABC):
    """Abstract base class for file types."""

    extension: ClassVar[str | frozenset[str] | None] = None

    def __init__(self):
        raise TypeError(f"{self.__class__.__name__} should not be instantiated")

    @classmethod
    def matches(cls, file_path: Path) -> bool:
        """Check if the file path matches the expected file type.

        Args:
            file_path: Path to the file.

        Returns:
            True if the file path matches the expected file type, False otherwise.
        """

        matching_extensions = cls.extension if isinstance(cls.extension, set) else {cls.extension}
        return any("".join(file_path.suffixes) == ext for ext in matching_extensions)

    @classmethod
    @abc.abstractmethod
    def validate(cls, contents: Any) -> None:
        """Validate the contents of a file.

        Args:
            contents: Content to validate.

        Raises:
            Exception: If the contents are invalid.
        """
        pass

    @classmethod
    @abc.abstractmethod
    def write(cls, file_path: Path, contents: Any) -> None:
        """Write content to a file.

        Args:
            file_path: Path to the file.
            contents: Content to write to the file.
        """
        pass



from pathlib import Path
from typing import Any, ClassVar

from .base import FileType


class TextFile(FileType):
    """A class for validating and writing text files.

    Examples:
        >>> with tempfile.NamedTemporaryFile() as tmp_file:
        ...     fp = Path(tmp_file.name)
        ...     TextFile.write(fp, "Hello, World!")
        ...     fp.read_text()
        'Hello, World!'

    `TextFile` can only accept string contents.

        >>> TextFile.validate("Hello!")
        >>> TextFile.validate(123)
        Traceback (most recent call last):
            ...
        ValueError: Contents must be a string; got <class 'int'>
    """

    extension: ClassVar[str] = ".txt"

    @classmethod
    def validate(cls, contents: Any):
        if not isinstance(contents, str):
            raise ValueError(f"Contents must be a string; got {type(contents)}")

    @classmethod
    def write(cls, file_path: Path, contents: Any) -> None:
        file_path.write_text(contents)



import importlib.metadata
import logging
from collections import defaultdict
from pathlib import Path

from .file_types.base import FileType

logger = logging.getLogger(__name__)


class FileTypeMatcher:
    """A class to match file types based on their extensions.

    This class checks the python entry points to identify all registered file types and loads them, then
    dynamically matches files specified via getitem to appropriate file type.
    """

    def __init__(self):
        eps = importlib.metadata.entry_points(group="yaml_to_disk.file_types")

        file_types = defaultdict(list)

        for name in eps.names:
            try:
                file_type = eps[name].load()
            except Exception as e:
                raise ImportError(f"Failed to load entry point {name}") from e

            if not issubclass(file_type, FileType):
                raise TypeError(f"Entry point {name} does not subclass FileType")
            if file_type.extension is None:
                raise ValueError(f"FileType {name} has no extension defined")
            if isinstance(file_type.extension, set | frozenset):
                for ext in file_type.extension:
                    file_types[ext].append(file_type)
            else:
                file_types[file_type.extension].append(file_type)

        err_lines = [f"{ext}: {types}" for ext, types in file_types.items() if len(types) > 1]
        if err_lines:
            raise ValueError(
                f"Multiple file types registered for the same extension:\n{', '.join(err_lines)}"
            )

        self.file_types = {k: v[0] for k, v in file_types.items()}

    def __call__(self, key: str | Path) -> FileType:
        match key:
            case Path() as fp:
                key_suffix = "".join(fp.suffixes)
                key_path = fp
            case str() as suffix:
                key_suffix = suffix
                key_path = Path(suffix)
            case _:
                raise TypeError(f"Key must be a string or Path; got {type(key)}")

        if key_suffix in self.file_types:
            return self.file_types[key_suffix]

        matching_types = []
        for ext, file_type in self.file_types.items():
            if file_type.matches(key_path):
                logger.debug(f"File type {file_type.__name__} matches {key_path} though ext {ext} doesn't")
                matching_types.append(file_type)

        if len(matching_types) == 0:
            raise ValueError(f"No file type found for {key_suffix}")
        elif len(matching_types) == 1:
            return matching_types[0]
        else:
            raise ValueError(
                f"Multiple file types found for {key_suffix}: "
                f"{', '.join([type.__name__ for type in matching_types])}"
            )


FILE_TYPE_MATCHER = FileTypeMatcher()
